@@ -14,18 +14,13 @@
 #  limitations under the License.
 
 import sys
-import elements
-import blockdiag.DiagramMetrics
+from seqdiag import elements
+import blockdiag.metrics
 from blockdiag.utils import Box, XY
-from blockdiag.utils.collections import namedtuple
-
-try:
-    from blockdiag.utils.PILTextFolder import PILTextFolder as TextFolder
-except ImportError:
-    from blockdiag.utils.TextFolder import TextFolder
+from blockdiag.utils.collections import namedtuple, defaultdict
 
 
-class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
+class DiagramMetrics(blockdiag.metrics.DiagramMetrics):
     edge_height = 10
 
     def __init__(self, diagram, **kwargs):
@@ -61,17 +56,59 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
             font = self.font_for(edge)
             if edge.leftnote:
                 edge.leftnotesize = self.textsize(edge.leftnote, font=font)
-                if height < edge.leftnotesize.y:
-                    height = edge.leftnotesize.y
+                if height < edge.leftnotesize.height:
+                    height = edge.leftnotesize.height
 
             if edge.rightnote:
                 edge.rightnotesize = self.textsize(edge.rightnote, font=font)
-                if height < edge.rightnotesize.y:
-                    height = edge.rightnotesize.y
+                if height < edge.rightnotesize.height:
+                    height = edge.rightnotesize.height
 
             self.spreadsheet.set_node_height(edge.order + 1, height)
             self.expand_pagesize_for_note(edge)
 
+        span_width = defaultdict(int)
+        span_height = defaultdict(int)
+        for block in diagram.altblocks:
+            x1, y1 = block.xy
+            x2 = x1 + block.colwidth
+            y2 = y1 + block.colheight
+
+            for y in range(y1, y2):
+                span_width[(x1, y)] += 1
+                span_width[(x2, y)] += 1
+
+            for x in range(x1, x2):
+                if block.type != 'else':
+                    span_height[(x, y1)] += 1
+                span_height[(x, y2)] += 1
+
+        for x in range(self.node_count + 1):
+            widths = [span_width[xy] for xy in span_width if xy[0] == x]
+            if widths:
+                width = self.span_width + max(widths) * self.cellsize
+                self.spreadsheet.set_span_width(x, width)
+
+        for y in range(0, len(self.edges) + 1):
+            blocks = [b for b in diagram.altblocks if b.edges[0].order == y]
+            span_height = self.spreadsheet.span_height[y]
+            span_height = 0
+
+            if blocks:
+                max_ylevel_top = max(b.ylevel_top for b in blocks)
+                span_height = (self.spreadsheet.span_height[y + 1] +
+                               self.cellsize * 5 / 2 * (max_ylevel_top - 1) +
+                               self.cellsize)
+                self.spreadsheet.set_span_height(y + 1, span_height)
+
+            blocks = [b for b in diagram.altblocks if b.edges[-1].order == y]
+            if blocks:
+                max_ylevel_bottom = max(b.ylevel_bottom for b in blocks)
+                span_height = (self.spreadsheet.span_height[y + 2] +
+                               self.cellsize / 2 * (max_ylevel_bottom - 1))
+
+                self.spreadsheet.set_span_height(y + 2, span_height)
+
     def pagesize(self, width=None, height=None):
         width = self.node_count
         height = len(self.edges) + len(self.separators) + 1
@@ -82,7 +119,8 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
         height = len(self.edges) + len(self.separators)
         dummy = elements.DiagramNode(None)
         dummy.xy = XY(1, height)
-        x, y = self.spreadsheet._node_bottomright(dummy, use_padding=False)
+        _, y = self.spreadsheet._node_bottomright(dummy, use_padding=False)
+        y += self.spreadsheet.span_height[len(self.edges) + 1] / 2
         return y
 
     @property
@@ -113,17 +151,20 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
         return lines
 
     def activity_box(self, node, activity):
+        edges = dict([e.order, e] for e in self.edges + self.separators)
+
         # y coodinates for top of activity box
         starts = activity['lifetime'][0]
-        edge = self.edges[starts]
+        edge = edges[starts]
         y1 = self.edge(edge).baseheight
-        if edge.diagonal and edge.node2 == node:
-            y1 += self.edge_height * 3 / 4
+        if isinstance(edge, elements.DiagramEdge):
+            if edge.diagonal and edge.node2 == node:
+                y1 += self.edge_height * 3 / 4
 
         # y coodinates for bottom of activity box
         ends = activity['lifetime'][-1] + 1
-        if ends < len(self.edges):
-            y2 = self.edge(self.edges[ends]).baseheight
+        if ends < len(edges):
+            y2 = self.edge(edges[ends]).baseheight
         else:
             y2 = self.bottomheight + self.cellsize * 2
 
@@ -137,8 +178,10 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
     def activity_shadow(self, node, activity):
         box = self.activity_box(node, activity)
 
-        return (box[0] + self.shadow_offset.x, box[1] + self.shadow_offset.y,
-                box[2] + self.shadow_offset.x, box[3] + self.shadow_offset.y)
+        return Box(box[0] + self.shadow_offset.x,
+                   box[1] + self.shadow_offset.y,
+                   box[2] + self.shadow_offset.x,
+                   box[3] + self.shadow_offset.y)
 
     def edge_textsize(self, edge):
         width = 0
@@ -148,9 +191,10 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
                 cell = self.cell(edge.node1)
                 width = self.edge(edge).right - cell.center.x
             else:
-                width = self.cell(edge.right_node).center.x - \
-                        self.cell(edge.left_node).center.x
-            width, height = self.textsize(edge.label, width,
+                width = (self.cell(edge.right_node).center.x -
+                         self.cell(edge.left_node).center.x -
+                         self.cellsize * 4)  # 4: width of activity and padding
+            width, height = self.textsize(edge.label, width=width,
                                           font=self.font_for(edge))
 
         return XY(width, height)
@@ -160,8 +204,8 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
             cell = self.cell(edge.left_node)
             width = cell.center.x - self.cellsize * 6
 
-            if width < edge.leftnotesize.x:
-                span_width = edge.leftnotesize.x - width
+            if width < edge.leftnotesize.width:
+                span_width = edge.leftnotesize.width - width
                 self.spreadsheet.span_width[0] += span_width
 
         if edge.rightnote:
@@ -175,8 +219,8 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
             if edge.right_node.xy.x + 1 == self.node_count:
                 width -= self.cellsize * 2
 
-            if width < edge.rightnotesize.x:
-                span_width = edge.rightnotesize.x - width
+            if width < edge.rightnotesize.width:
+                span_width = edge.rightnotesize.width - width
                 self.spreadsheet.span_width[self.node_count] += span_width
 
     def cell(self, obj, use_padding=True):
@@ -184,16 +228,45 @@ class DiagramMetrics(blockdiag.DiagramMetrics.DiagramMetrics):
             klass = namedtuple('_', 'xy width height colwidth colheight')
             edge = klass(XY(1, obj.order + 1), None, None, 1, 1)
             return super(DiagramMetrics, self).cell(edge, use_padding=False)
+        elif isinstance(obj, elements.AltBlock):
+            box = super(DiagramMetrics, self).cell(obj, use_padding=False)
+            return AltBlockMetrics(self, obj, box)
         else:
             return super(DiagramMetrics, self).cell(obj, use_padding)
 
     def edge(self, edge):
-        return EdgeMetrics(edge, self)
+        if isinstance(edge, elements.EdgeSeparator):
+            return self.separator(edge)
+        else:
+            return EdgeMetrics(edge, self)
 
     def separator(self, separator):
         return SeparatorMetrics(separator, self)
 
 
+class AltBlockMetrics(blockdiag.metrics.NodeMetrics):
+    def __init__(self, metrics, block, box):
+        self.block = block
+        sheet = metrics.spreadsheet
+        cellsize = metrics.cellsize
+
+        box[0] -= (sheet.span_width[block.xy.x + 1] / 2 -
+                   cellsize * (block.xlevel - 1))
+        box[1] -= cellsize * 5 / 2 * (block.ylevel_top - 1) + cellsize * 3
+        box[2] += (sheet.span_width[block.xy.x + block.colwidth + 2] / 2 -
+                   cellsize * (block.xlevel - 1))
+        box[3] += cellsize * (block.ylevel_bottom - 1) + cellsize
+
+        super(AltBlockMetrics, self).__init__(metrics, *box)
+
+    @property
+    def textbox(self):
+        size = self.metrics.textsize(self.block.type,
+                                     font=self.metrics.font_for(self.block))
+        return Box(self.x1, self.y1,
+                   self.x1 + size.width, self.y1 + size.height)
+
+
 class EdgeMetrics(object):
     def __init__(self, edge, metrics):
         self.metrics = metrics
@@ -202,8 +275,8 @@ class EdgeMetrics(object):
     @property
     def baseheight(self):
         cell = self.metrics.cell(self.edge)
-        if cell.height == self.edge.leftnotesize.y or \
-           cell.height == self.edge.rightnotesize.y:
+        if (cell.height == self.edge.leftnotesize.height or
+           (cell.height == self.edge.rightnotesize.height)):
             return cell.center.y
         else:
             return cell.top.y + self.edge.textheight
@@ -223,7 +296,7 @@ class EdgeMetrics(object):
 
             x = cell.bottom.x + width
         else:
-            x = cell.bottom.x - m.cell.cellsize
+            x = cell.bottom.x - m.cellsize
 
             if self.edge.failed:
                 x -= self.metrics.edge_length / 2
@@ -249,8 +322,8 @@ class EdgeMetrics(object):
                     XY(x2, baseheight + fold_height),
                     XY(x1 + m.cellsize, baseheight + fold_height)]
         else:
-            x1 = self.metrics.node(self.edge.left_node).bottom.x + \
-                 self.activity_line_width(self.edge.left_node)
+            x1 = (self.metrics.node(self.edge.left_node).bottom.x +
+                  self.activity_line_width(self.edge.left_node))
             x2 = self.metrics.node(self.edge.right_node).bottom.x
 
             margin = m.cellsize
@@ -329,7 +402,8 @@ class EdgeMetrics(object):
                 self.activity_line_width(self.edge.node1)
         elif self.edge.direction == 'right':
             x = m.node(self.edge.left_node).bottom.x + \
-                self.activity_line_width(self.edge.node1)
+                self.activity_line_width(self.edge.left_node) + \
+                self.metrics.cellsize / 2
         else:  # left
             x = m.node(self.edge.right_node).bottom.x - self.edge.textwidth
 
@@ -357,9 +431,13 @@ class EdgeMetrics(object):
         cell = m.cell(self.edge.left_node)
         notesize = self.edge.leftnotesize
 
-        x = cell.center.x - m.cellsize * 3 - notesize.x
-        y = self.baseheight - notesize.y / 2
-        return Box(x, y, x + notesize.x, y + notesize.y)
+        x = cell.center.x - m.cellsize * 3 - notesize.width
+        y = self.baseheight - notesize.height / 2
+
+        if self.edge.failed and self.edge.direction == 'left':
+            x += self.metrics.edge_length / 2 - m.cellsize
+
+        return Box(x, y, x + notesize.width, y + notesize.height)
 
     @property
     def leftnoteshape(self):
@@ -381,12 +459,14 @@ class EdgeMetrics(object):
         cell = m.cell(self.edge.right_node)
         if self.edge.direction == 'self':
             x = self.right + m.cellsize * 2
+        elif self.edge.failed and self.edge.direction == 'right':
+            x = self.right + m.cellsize * 4
         else:
             x = cell.center.x + m.cellsize * 2
 
         notesize = self.edge.rightnotesize
-        y = self.baseheight - notesize.y / 2
-        return Box(x, y, x + notesize.x, y + notesize.y)
+        y = self.baseheight - notesize.height / 2
+        return Box(x, y, x + notesize.width, y + notesize.height)
 
     @property
     def rightnoteshape(self):
@@ -410,10 +490,14 @@ class SeparatorMetrics(object):
         y2 = y1 + self.metrics.node_height
         d = self.metrics.cellsize / 4
 
-        lines = TextFolder((x1, y1, x2, y2), separator.label,
-                           metrics.font_for(self))
-        box = lines.outlinebox
-        self.labelbox = (box[0] - d, box[1] - d, box[2] + d, box[3] + d)
+        font = metrics.font_for(self)
+        size = metrics.textsize(separator.label, font, x2 - x1)
+        dx = (x2 - x1 - size.width) / 2
+        dy = (y2 - y1 - size.height) / 2
+        self.labelbox = Box(x1 + dx - d,
+                            y1 + dy - d,
+                            x1 + dx + size.width + d,
+                            y1 + dy + size.height + d)
 
     @property
     def baseheight(self):
